"""Interface for uploading pricing models data."""

import os
from typing import List, Optional, Union
from rmlab_errors import raise_from_list
from rmlab._api.upload import APIUploadInternal
from rmlab._data.enums import (
    CustomersModelKind,
    ParametricModelKind,
)


class APIUploadCustomersModels(APIUploadInternal):
    """Exposes functions for uploading customers models data to the server."""

    async def upload_customers_request_model(
        self,
        scen_id: int,
        content: Union[str, dict],
    ) -> None:
        """Upload to server a parametric model defined in file, modelling how customers request books.

        Args:
            scen_id (int): Scenario ID
            content (str): Request model as file or json dict (require `id` identifier field if the latter)
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.CUSTOMERS,
            kind=CustomersModelKind.REQUEST,
            content=content,
        )

    async def upload_customers_choice_model(
        self,
        scen_id: int,
        content: Union[str, dict],
    ) -> None:
        """Upload to server a parametric model defined in file, modelling how customers choose between competing book offers.

        Args:
            scen_id (int): Scenario ID
            content (str): Request model as file or json dict (require `id` identifier field if the latter)
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.CUSTOMERS,
            kind=CustomersModelKind.CHOICE,
            content=content,
        )

    async def upload_batch_customers_models(
        self,
        scen_id: int,
        *,
        request_models: Optional[List[Union[str, dict]]] = None,
        choice_models: Optional[List[Union[str, dict]]] = None,
    ):
        if request_models is None:
            request_models = list()
        if choice_models is None:
            request_models = list()

        not_found_errors = \
            [FileNotFoundError(fn) for fn in request_models if isinstance(fn, str) and not os.path.exists(fn)] + \
            [FileNotFoundError(fn) for fn in choice_models if isinstance(fn, str) and not os.path.exists(fn)]
        
        type_errors = \
            [TypeError(cnt) for cnt in request_models if (not isinstance(cnt, dict)) and (not isinstance(cnt, str))] + \
            [TypeError(cnt) for cnt in choice_models if not isinstance(cnt, dict) and (not isinstance(cnt, str))]

        raise_from_list(not_found_errors + type_errors)

        for crm in request_models:

            await self._upload_parametric_model(
                scen_id, ParametricModelKind.CUSTOMERS, CustomersModelKind.REQUEST, crm
            )

        for ccm in choice_models:

            await self._upload_parametric_model(
                scen_id, ParametricModelKind.CUSTOMERS, CustomersModelKind.CHOICE, ccm
            )



from rmlab._api.base import APIBaseInternal

class APIPersistence(APIBaseInternal):
    """Interface to interact with data snapshots"""

    async def save_snapshot(self, snapshot_name: str):
        return await self._submit_call("api-snapshot-save", snapshot_name=snapshot_name)

    async def restore_snapshot(self, snapshot_name: str):
        return await self._submit_call("api-snapshot-restore", snapshot_name=snapshot_name)

    async def delete_snapshot(self, snapshot_name: str):
        return await self._submit_call("api-snapshot-delete", snapshot_name=snapshot_name)
    
    async def list_snapshots(self):
        return await self._submit_call("api-snapshot-list")


"""A pricing behavior defines the strategy under which flights assign prices to seats."""

import json
from typing import Union
from dataclasses import dataclass
from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import FileExtensions, PricingModelKind
from rmlab.data.parametric._pricing_base import PricingModel


@dataclass
class BaseBehaviorModel(PricingModel):
    """Base class for all types of pricing behavior models."""

    pass

@dataclass
class RakeBehaviorModel(BaseBehaviorModel):
    """A kind of pricing behavior model in which a set of seat thresholds are used to allocate seats to fares based on the flight occupation.
    These thresholds can be squeezed with respect to the total flight occupation or shifted up from zero occupation.

    Args:
      zero_shift (float): Per-unit relative displacement of all the thresholds
      squeeze (float): Per-unit relative contraction of the all thresholds

    """

    zero_shift: float
    squeeze: float

@dataclass
class RakeStraightBehaviorModel(RakeBehaviorModel):
    """A kind of pricing behavior model in which a set of seat thresholds are used to allocate seats to fares based on the flight occupation.
    These thresholds are straight, in the sense that the seat limits form a straight line with day before departure.

    Args:
      seats_per_dbd (float): Rate of increase in number of seats per day of all thresholds

    Example
    ```py
    behavior = RakeStraightBehaviorModel(seats_per_dbd=0.15, zero_shift=0.06, squeeze=0.7)
    ```
    """

    seats_per_dbd: float

@dataclass
class RakeEBCBehaviorModel(RakeStraightBehaviorModel):
    """A kind of pricing behavior model in which a set of seat thresholds are derived from an expected booking curve and a confidence level.
    Derives from RakeStraightBehavior to have default rakes when there is no input data from which the EBC is built.
    Args:
        confidence (float): Level from which the thresholds slope is derived.
    Example
    ```py
    behavior = ExpectedBookingCurveThresholds(confidence=0.95, seats_per_dbd=0.15, zero_shift=0.06, squeeze=0.7)
    ```
    """

    confidence: float


def make_pricing_behavior_from_json(
    filename_or_dict: Union[str, dict]
) -> BaseBehaviorModel:
    """Make a pricing behavior instance from a json representation (from file or dict).

    Args:
        filename_or_dict (Union[str, dict]): JSON filename or dictionary in json format

    Examples from dict:
    ```py
    dict_rake_straight_behavior = dic()
    dict_rake_straight_behavior["type"] = "rake_straight"
    dict_rake_straight_behavior["seats_per_dbd"] = 0.15
    dict_rake_straight_behavior["zero_shift"] = 0.06
    dict_rake_straight_behavior["squeeze"] = 0.7
    my_pricing_rake_straight_behavior = make_from_json(dict_rake_straight_behavior)
    ```

    Example from file:
    `my_pricing_rake_straight_behavior.json`
    ```json
    {
      "type" : "rake_straight",
      "seats_per_dbd" : 0.15,
      "zero_shift" : 0.06,
      "squeeze" : 0.7
    }
    ```

    ```py
    my_pricing_rake_straight_behavior = make_pricing_behavior_from_json("my_pricing_behavior.json")
    ```
    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_dict)

    if not isinstance(content, dict):
        raise TypeError(f"Expected dict format in {filename_or_dict}")

    content: dict

    if content["type"] == "rake_straight":

        return RakeStraightBehaviorModel(
            id=model_id,
            filename=filename,
            cnt=json.dumps(content),
            hash=md5hash,
            kind=PricingModelKind.BEHAVIOR,
            extension=FileExtensions.JSON,
            seats_per_dbd=float(content["seats_per_dbd"]),
            zero_shift=float(content["zero_shift"]),
            squeeze=float(content["squeeze"]),
        )

    elif content["type"] == "rake_ebc":

        return RakeEBCBehaviorModel(
            id=model_id,
            filename=filename,
            cnt=json.dumps(content),
            hash=md5hash,
            kind=PricingModelKind.BEHAVIOR,
            extension=FileExtensions.JSON,
            seats_per_dbd=float(content["seats_per_dbd"]),
            confidence=float(content["confidence"]),
            zero_shift=float(content["zero_shift"]),
            squeeze=float(content["squeeze"]),
        )

    else:

        raise ValueError(f'Unknown pricing behavior type `{content["type"]}`')



from typing import Union
from rmlab._api.upload import APIUploadInternal
from rmlab.data.parametric.filter import PFilter


class APIUploadParametric(APIUploadInternal):
    """Exposes functions for uploading parametric filters to the server."""

    async def upload_parametric_filters(self, scen_id: int, filters: Union[str, list]) -> None:
        """Upload a set of parametric filters defined in a file.

        ```json
        {"json": "example"}
        ```

        Args:
            scen_id (int): Scenario ID
            filters: Parametric filters as file (csv or json) or as json list

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=PFilter, items=filters
        )



"""
This script provides dataclasses storing arrays representing temporal data of flights.
"""
from dataclasses import dataclass
from datetime import datetime, timezone
from itertools import accumulate
from typing import List, Mapping

from rmlab._data.enums import FlightEvent

_MSecFactor = 1000

@dataclass
class FlightData:
    """Base dataclass for derived classes holding time series data associated to a flight.

    Args:
        id (str): Flight ID to which time series data arrays are associated
        timestamps_array (List[datetime]): Sequence of time stamps with milliseconds granularity
    """

    id: str
    timestamps_array: List[datetime]

    def __post_init__(self):
        if not isinstance(self.timestamps_array, list):
            raise TypeError(f"Expected `list` type in timestamps array. Got `{type(self.timestamps_array)}`")
        if len(self.timestamps_array) > 0:
            if isinstance(self.timestamps_array[0], int):
                # NOTE Assumed input timestamps are in millisec
                self.timestamps_array = [
                    datetime.fromtimestamp(ts / _MSecFactor, tz=timezone.utc) for ts in self.timestamps_array
                ]
            elif not isinstance(self.timestamps_array[0], datetime):
                raise TypeError(f"Expected `datetime` or `int` time for timestamps. Got `{type(self.timestamps_array[0])}`")
            # else timestamps are already a datetime, do nothing
        # else empty array

    def id_serial_timestamps(self):

        if not isinstance(self.timestamps_array, list):
            raise TypeError(f"Expected `list` type in timestamps array. Got `{type(self.timestamps_array)}`")
        if len(self.timestamps_array) > 0:
            if isinstance(self.timestamps_array[0], int):
                return self.id, {"Timestamp": self.timestamps_array}
            elif isinstance(self.timestamps_array[0], datetime):
                return self.id, {"Timestamp": [round(ts.timestamp() * _MSecFactor) for ts in self.timestamps_array]}
            else:
                raise TypeError(f"Expected `datetime` or `int` time for timestamps. Got `{type(self.timestamps_array[0])}`")
        # else empty array

    def is_empty(self):
        return len(self.timestamps_array) == 0


@dataclass
class FlightDataBooks(FlightData):
    """Data class with arrays related to books associated to a flight

    Args:
        fares_array (List[str]): Stores the enabled fares at each time stamp
        pps_array (List[int]): Stores the price per seat associated to each booking in cents
        seats_array (List[int]): Stores the seats booked for each booking
        cumulated_seats_array (List[int]): Stores the cumulated booked seats at each time stamp
        cumulated_revenue_array (List[int]): Stores the cumulated revenue at each time stamp in cents
    """

    fares_array: List[str]
    pps_array: List[int]
    seats_array: List[int]
    cumulated_seats_array: List[int]
    cumulated_revenue_array: List[int]

    def __post_init__(self):
        super().__post_init__()
        if self.cumulated_seats_array is None:
            self.cumulated_seats_array = list(accumulate(self.seats_array))
        if self.cumulated_revenue_array is None:
            self.cumulated_revenue_array = list(accumulate([pps*s for pps, s in zip(self.pps_array, self.seats_array)]))

        assert len(self.fares_array) == len(self.timestamps_array)
        assert len(self.pps_array) == len(self.timestamps_array)
        assert len(self.seats_array) == len(self.timestamps_array)
        assert len(self.cumulated_revenue_array) == len(self.timestamps_array)
        assert len(self.cumulated_seats_array) == len(self.timestamps_array)

    def id_dict(self):
        id, d = self.id_serial_timestamps()
        return id, {**d,
                    "Seats": self.seats_array,
                    "PPS": self.pps_array,
                    "Fares": self.fares_array}


@dataclass
class FlightDataForecastedBooks(FlightData):
    """Data class with arrays related to forecasted books associated to a flight

    Args:
        pps_array (List[float]): Stores the forecasted price per seat at each time stamp
        seats_array (List[float]): Stores the forecasted cumulated booked seats at each time stamp
    """

    pps_array: List[float]
    seats_array: List[float]

    def __post_init__(self):
        super().__post_init__()
        assert len(self.pps_array) == len(self.timestamps_array)
        assert len(self.seats_array) == len(self.timestamps_array)


@dataclass
class FlightDataThresholdSettings(FlightData):
    """Data class with arrays of seat thresholds for each fare associated to a flight

    Args:
        fare_to_threshold_array (Mapping[str], List[int]): Maps each fare identifier
            to an array storing the seat thresholds at each time stamp.
    """

    fare_to_threshold_array: Mapping[str, List[int]]

    def id_dict(self):
        id, d = self.id_serial_timestamps()
        return id, {**d, **self.fare_to_threshold_array}

    def __post_init__(self):
        super().__post_init__()
        assert all([len(threshold) == len(self.timestamps_array)
                    for threshold in self.fare_to_threshold_array.values()])


@dataclass
class FlightDataPricePerSeatSettings(FlightData):
    """Data class with arrays of prices per seat for each fare associated to a flight

    Args:
        fare_to_pps_array (Mapping[str], List[int]): Maps each fare identifier
            to an array storing the prices per seat at each time stamp.
    """

    fare_to_pps_array: Mapping[str, List[int]]

    def id_dict(self):
        id, d = self.id_serial_timestamps()
        return id, {**d, **self.fare_to_pps_array}

    def __post_init__(self):
        super().__post_init__()
        assert all([len(pps) == len(self.timestamps_array)
                    for pps in self.fare_to_pps_array.values()])


@dataclass
class FlightDataEvents(FlightData):
    """Data class holding a sequence of events associated to a flight

    Args:
        events_array (List[FlightEvent]): Stores the events at each time stamp.
    """

    events_array: List[FlightEvent]

    def __post_init__(self):
        super().__post_init__()
        assert len(self.events_array) == len(self.timestamps_array)


@dataclass
class FlightDataQForecast(FlightData):
    """Data class with arrays result of q-forecast.

    Args:
        frat5_array (List[float]): Array of FRAT5 values
        book_qequivalent_array: Array of Q-equivalent books.
    """
    frat5_array: List[float]
    book_qequivalent_array: List[float]

    def __post_init__(self):
        super().__post_init__()
        assert len(self.frat5_array) == len(self.timestamps_array)
        assert len(self.book_qequivalent_array) == len(self.timestamps_array)



"""Interface for uploading customers models data."""

import os
from typing import List, Optional, Union
from rmlab_errors import raise_from_list
from rmlab._api.upload import APIUploadInternal
from rmlab._data.enums import (
    ParametricModelKind,
    PricingModelKind,
)


class APIUploadPricingModels(APIUploadInternal):
    """Exposes functions for uploading pricing models data to the server."""

    async def upload_pricing_range_model(
        self,
        scen_id: int,
        content: Union[str, dict],
    ) -> None:
        """Upload to server a parametric model defined in file, specifying a pricing range to be applied by flights.

        Args:
            scen_id (int): Scenario ID
            content: Model content as file or json dict (require `id` identifier field if the latter)
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.PRICING,
            kind=PricingModelKind.RANGE,
            content=content,
        )

    async def upload_pricing_behavior_model(
        self,
        scen_id: int,
        content: Union[str, dict],
    ) -> None:
        """Upload to server a parametric model defined in file, specifying the pricing behavior/strategy
        under which flights assign prices to seats.

        Args:
            scen_id (int): Scenario ID
            content: Model content as file or json dict (require `id` identifier field if the latter)
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.PRICING,
            kind=PricingModelKind.BEHAVIOR,
            content=content,
        )

    async def upload_pricing_optimizer_model(
        self,
        scen_id: int,
        content: Union[str, dict],
    ) -> None:
        """Upload to server a parametric model defined in file, specifying the pricing optimization methods
        under which flights operate, to adapt the prices given current and past demand.

        Args:
            scen_id (int): Scenario ID
            content: Model content as file or json dict (require `id` identifier field if the latter)
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.PRICING,
            kind=PricingModelKind.OPTIMIZER,
            content=content,
        )

    async def upload_batch_pricing_models(
        self,
        scen_id: int,
        *,
        range_models: Optional[List[Union[str, dict]]] = None,
        behavior_models: Optional[List[Union[str, dict]]] = None,
        optimizer_models: Optional[List[Union[str, dict]]] = None,
    ):
        if range_models is None:
            range_models = list()
        if behavior_models is None:
            behavior_models = list()
        if optimizer_models is None:
            optimizer_models = list()

        not_existing_filenames = \
            [FileNotFoundError(fn) for fn in range_models if isinstance(fn, str) and not os.path.exists(fn)] + \
            [FileNotFoundError(fn) for fn in behavior_models if isinstance(fn, str) and not os.path.exists(fn)] + \
            [FileNotFoundError(fn) for fn in optimizer_models if isinstance(fn, str) and not os.path.exists(fn)]

        wrong_types = \
            [TypeError("Range model expects dict or file name") for cnt in range_models if (not isinstance(cnt, dict)) and (not isinstance(cnt, str))] + \
            [TypeError("Behavior model expects dict or file name") for cnt in behavior_models if (not isinstance(cnt, dict)) and (not isinstance(cnt, str))] + \
            [TypeError("Optimizer model expects dict or file name") for cnt in optimizer_models if (not isinstance(cnt, dict)) and (not isinstance(cnt, str))]

        raise_from_list(not_existing_filenames + wrong_types)

        for prm in range_models:
            await self._upload_parametric_model(
                scen_id, ParametricModelKind.PRICING, PricingModelKind.RANGE, prm
            )

        for pbm in behavior_models:
            await self._upload_parametric_model(
                scen_id, ParametricModelKind.PRICING, PricingModelKind.BEHAVIOR, pbm
            )

        for pom in optimizer_models:
            await self._upload_parametric_model(
                scen_id, ParametricModelKind.PRICING, PricingModelKind.OPTIMIZER, pom
            )



"""Interface for fetching data related to parametric models and filters."""

from typing import List
from rmlab.data.items import PModel
from rmlab.data.parametric.filter import PFilter
from rmlab._api.fetch import APIFetchInternal
from rmlab._data.enums import CustomersModelKind, ParametricModelKind, PricingModelKind
from rmlab.data.parametric.customers_request import make_customers_request_model_from_json, BaseRequestModel
from rmlab.data.parametric.customers_choice import make_customers_choice_model_from_json, BaseChoiceModel

from rmlab.data.parametric.pricing_range import make_pricing_range_from_json, RangeModel
from rmlab.data.parametric.pricing_behavior import make_pricing_behavior_from_json, BaseBehaviorModel
from rmlab.data.parametric.pricing_optimizer import make_pricing_optimizer_from_json, OptimizerModel



class APIFetchParametric(APIFetchInternal):
    """Exposes functions for fetching flight data from the server."""

    async def fetch_parametric_filters(self, scen_id: int) -> List[PFilter]:
        """Fetch a list of all parametric filters of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of parametric filters
        """

        return await self._fetch_bounded_items(scen_id, PFilter)

    async def fetch_parametric_models(self, scen_id: int) -> List[PModel]:
        """Fetch a list of all parametric models of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of parametric models
        """

        return await self._fetch_bounded_items(scen_id, PModel)
    
    async def fetch_customers_request_model(self, scen_id: int, pmodel_id: str) -> BaseRequestModel:
        """Fetch a customers request model from id in scenario
        
        Args:
            scen_id (int): Scenario ID
            pmodel_id (str): Model ID
        
        Returns:
            The requested customers request model
        """

        json = await self._fetch_pmodel(scen_id,
                                  f"{ParametricModelKind.CUSTOMERS.value}_{CustomersModelKind.REQUEST.value}",
                                  pmodel_id)
        
        model = make_customers_request_model_from_json(json)
        model.id = pmodel_id
        return model

    async def fetch_customers_choice_model(self, scen_id: int, pmodel_id: str) -> BaseChoiceModel:
        """Fetch a customers choice model from id in scenario
        
        Args:
            scen_id (int): Scenario ID
            pmodel_id (str): Model ID
        
        Returns:
            The requested customers choice model
        """

        json = await self._fetch_pmodel(scen_id,
                                  f"{ParametricModelKind.CUSTOMERS.value}_{CustomersModelKind.CHOICE.value}",
                                  pmodel_id)
        model = make_customers_choice_model_from_json(json)
        model.id = pmodel_id
        return model

    async def fetch_pricing_range_model(self, scen_id: int, pmodel_id: str) -> RangeModel:
        """Fetch a pricing range model from id in scenario
        
        Args:
            scen_id (int): Scenario ID
            pmodel_id (str): Model ID
        
        Returns:
            The requested pricing range model
        """

        json = await self._fetch_pmodel(scen_id,
                                  f"{ParametricModelKind.PRICING.value}_{PricingModelKind.RANGE.value}",
                                  pmodel_id)
        model = make_pricing_range_from_json(json)
        model.id = pmodel_id
        return model

    async def fetch_pricing_behavior_model(self, scen_id: int, pmodel_id: str) -> BaseBehaviorModel:
        """Fetch a pricing behavior model from id in scenario
        
        Args:
            scen_id (int): Scenario ID
            pmodel_id (str): Model ID
        
        Returns:
            The requested pricing behavior model
        """

        json = await self._fetch_pmodel(scen_id,
                                  f"{ParametricModelKind.PRICING.value}_{PricingModelKind.BEHAVIOR.value}",
                                  pmodel_id)
        model = make_pricing_behavior_from_json(json)
        model.id = pmodel_id
        return model

    async def fetch_pricing_optimizer_model(self, scen_id: int, pmodel_id: str) -> OptimizerModel:
        """Fetch a pricing optimizer model from id in scenario
        
        Args:
            scen_id (int): Scenario ID
            pmodel_id (str): Model ID
        
        Returns:
            The requested pricing optimizer model
        """

        json = await self._fetch_pmodel(scen_id,
                                  f"{ParametricModelKind.PRICING.value}_{PricingModelKind.OPTIMIZER.value}",
                                  pmodel_id)
        model = make_pricing_optimizer_from_json(json)
        model.id = pmodel_id
        return model


"""Interface to trigger optimization passes."""

from datetime import datetime
from typing import Optional, List
from rmlab._api.base import APIBaseInternal
from rmlab._data.types import DateTimeMinFormat, DateFormat


class APIOptimization(APIBaseInternal):
    """Exposes functions to run and schedule optimization runs on a set of flights
    and to fetch flights related to optimization targets."""

    async def trigger_optimization_pass(
        self,
        scen_id: int,
        airline_id: str,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ):
        """Triggers an optimization pass on all flights of an airline belonging to a citysector or sector.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID
            citysector_id (Optional[str], optional): Citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"At least one of `citysector_id`, `sector_id` must be defined"
            )

        await self._submit_call(
            "api-operation-optimization-trigger",
            scen_id=scen_id,
            airline_id=airline_id,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def schedule_optimization_pass(
        self,
        scen_id: int,
        airline_id: str,
        date_time: datetime,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ):
        """Schedules an optimization pass on all flights of an airline belonging to a citysector or sector to be run at specific date and time.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID
            date_time (datetime): Date and time at which the optimization pass is triggered.
            citysector_id (Optional[str], optional): Citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"At least one of `citysector_id`, `sector_id` must be defined"
            )

        await self._submit_call(
            "api-operation-optimization-schedule",
            scen_id=scen_id,
            date_time=datetime.strftime(date_time, DateTimeMinFormat),
            airline_id=airline_id,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def fetch_optimization_input_flights(
            self,
            scen_id: int,
            flight_id: str,
            citysector_id: str) -> List[str]:
        """Returns all flights whose data is used as input for optimizing the input flight.

        Args:
            scen_id (int): Scenario ID.
            flight_id (str): Flight ID.
            citysector_id (str): Citysector ID of flight.
        """

        return await self._submit_call(
            "api-operation-optimization-input-ids",
            scen_id=scen_id,
            flight_id=flight_id,
            citysector_id=citysector_id)

    async def fetch_optimization_scheduled_flights(
            self,
            scen_id: int,
            airline_id: str,
            sector_id: str,
            date_start: datetime,
            date_end: datetime) -> List[str]:
        """Returns all flights of an airline in a sector scheduled for optimization passes in a date interval

        Args:
            scen_id (int): Scenario ID.
            airline_id (str): Airline ID.
            sector_id (str): Sector ID of flight.
            date_start (datetime): Start of date interval.
            date_end (datetime): End of date interval.
        """

        return await self._submit_call(
            "api-operation-optimization-scheduled-ids",
            scen_id=scen_id,
            airline_id=airline_id,
            sector_id=sector_id,
            ds_start=datetime.strftime(date_start, DateFormat),
            ds_end=datetime.strftime(date_end, DateFormat))

    async def set_current_date(
            self,
            scen_id: int,
            date_current: datetime):
        """Set current date.

        Args:
            scen_id (int): Scenario ID.
            date_current (datetime): Current date to set.
        """

        if not isinstance(date_current, datetime):
            raise TypeError(f"Expected `datetime` type, got `{type(date_current)}`")

        return await self._submit_call(
            "api-operation-set-date",
            scen_id=scen_id,
            ds_current=datetime.strftime(date_current, DateFormat))
